"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const vscode_languageserver_1 = require("vscode-languageserver");
const references_1 = require("../util/references");
const treeUtils_1 = require("../util/treeUtils");
class CodeLensProvider {
    constructor(connection, forest, imports) {
        this.imports = imports;
        this.handleCodeLensRequest = (param) => __awaiter(this, void 0, void 0, function* () {
            this.connection.console.info(`A code lens was requested for ${param.textDocument.uri}`);
            const codeLens = [];
            const tree = this.forest.getTree(param.textDocument.uri);
            if (tree) {
                codeLens.push(...this.getExposingCodeLenses(tree));
                codeLens.push(...this.getReferencesCodeLenses(tree, param.textDocument.uri));
                return codeLens;
            }
        });
        this.handleCodeLensResolveRequest = (param) => __awaiter(this, void 0, void 0, function* () {
            const codelens = param;
            const data = codelens.data;
            this.connection.console.info(`A code lens resolve was requested`);
            if (data.codeLensType) {
                switch (data.codeLensType) {
                    case "exposed":
                        codelens.command = data.exposed
                            ? vscode_languageserver_1.Command.create("exposed", "")
                            : vscode_languageserver_1.Command.create("local", "");
                        break;
                    case "referenceCounter":
                        codelens.command = vscode_languageserver_1.Command.create(data.references.length === 1
                            ? "1 reference"
                            : `${data.references.length} references`, "");
                        break;
                    default:
                        break;
                }
            }
            return codelens;
        });
        this.connection = connection;
        this.forest = forest;
        this.connection.onCodeLens(this.handleCodeLensRequest);
        this.connection.onCodeLensResolve(this.handleCodeLensResolveRequest);
    }
    createExposingCodeLens(node, nameNode, tree, isFunction) {
        const exposed = isFunction
            ? treeUtils_1.TreeUtils.isExposedFunction(tree, nameNode.text)
            : treeUtils_1.TreeUtils.isExposedTypeOrTypeAlias(tree, nameNode.text);
        return vscode_languageserver_1.CodeLens.create(vscode_languageserver_1.Range.create(vscode_languageserver_1.Position.create(node.startPosition.row, node.startPosition.column), vscode_languageserver_1.Position.create(node.endPosition.row, node.endPosition.column)), { codeLensType: "exposed", exposed });
    }
    createReferenceCodeLens(placementNode, nameNode, uri, tree) {
        const definitionNode = treeUtils_1.TreeUtils.findDefinitionNodeByReferencingNode(nameNode, uri, tree, this.imports);
        const references = references_1.References.find(definitionNode, this.forest, this.imports);
        let refLocations = [];
        if (references) {
            refLocations = references.map(a => vscode_languageserver_1.Location.create(a.uri, vscode_languageserver_1.Range.create(vscode_languageserver_1.Position.create(a.node.startPosition.row, a.node.startPosition.column), vscode_languageserver_1.Position.create(a.node.endPosition.row, a.node.endPosition.column))));
        }
        return vscode_languageserver_1.CodeLens.create(vscode_languageserver_1.Range.create(vscode_languageserver_1.Position.create(placementNode.startPosition.row, placementNode.startPosition.column), vscode_languageserver_1.Position.create(placementNode.endPosition.row, placementNode.endPosition.column)), {
            codeLensType: "referenceCounter",
            references: refLocations,
        });
    }
    getExposingCodeLenses(tree) {
        const codeLens = [];
        tree.rootNode.children.forEach(node => {
            if (node.type === "value_declaration") {
                const functionName = treeUtils_1.TreeUtils.getFunctionNameNodeFromDefinition(node);
                if (functionName) {
                    if (node.previousNamedSibling &&
                        node.previousNamedSibling.type === "type_annotation") {
                        codeLens.push(this.createExposingCodeLens(node.previousNamedSibling, functionName, tree, true));
                    }
                    else {
                        codeLens.push(this.createExposingCodeLens(node, functionName, tree, true));
                    }
                }
            }
            else if (node.type === "type_declaration" ||
                node.type === "type_alias_declaration") {
                const typeNode = treeUtils_1.TreeUtils.findFirstNamedChildOfType("upper_case_identifier", node);
                if (typeNode) {
                    codeLens.push(this.createExposingCodeLens(node, typeNode, tree, false));
                }
            }
        });
        return codeLens;
    }
    getReferencesCodeLenses(tree, uri) {
        const codeLens = [];
        tree.rootNode.children.forEach(node => {
            if (node.type === "type_declaration" ||
                node.type === "type_alias_declaration") {
                const typeNode = treeUtils_1.TreeUtils.findFirstNamedChildOfType("upper_case_identifier", node);
                if (typeNode) {
                    codeLens.push(this.createReferenceCodeLens(node, typeNode, uri, tree));
                }
            }
        });
        treeUtils_1.TreeUtils.descendantsOfType(tree.rootNode, "value_declaration").forEach(node => {
            const functionName = treeUtils_1.TreeUtils.getFunctionNameNodeFromDefinition(node);
            if (functionName) {
                if (node.previousNamedSibling &&
                    node.previousNamedSibling.type === "type_annotation") {
                    codeLens.push(this.createReferenceCodeLens(node.previousNamedSibling, functionName, uri, tree));
                }
                else {
                    codeLens.push(this.createReferenceCodeLens(node, functionName, uri, tree));
                }
            }
        });
        const moduleNameNode = treeUtils_1.TreeUtils.getModuleNameNode(tree);
        if (moduleNameNode && moduleNameNode.lastChild) {
            codeLens.push(this.createReferenceCodeLens(moduleNameNode, moduleNameNode.lastChild, uri, tree));
        }
        return codeLens;
    }
}
exports.CodeLensProvider = CodeLensProvider;
//# sourceMappingURL=codeLensProvider.js.map